//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// October 2000
//
#include <EpicProc.h>
#include <EpicProcState.h>

#include <Dal.h>	// SAS
#include <OalServer.h>
#include <CalServer.h>
#include <param.h>
#include <errstr.h>
#include <metatask.h>
#include <Proc.h>
#include <Evaluator.h>

#include <attcalc.h>     // Tasks
#include <atthkgen.h>
#include <badpixfind.h>
#include <evlistcomb.h>
#include <hkgtigen.h>
#include <timeappend.h>
#include <tabgtigen.h>
#include <gtimerge.h>
#include <gtialign.h>
#include <evselect.h>
#include <dscopyblock.h>
#include <dsrename.h>

#include <algorithm>
#include <cstdlib>
#include <unistd.h>

void Unlink(const string & f){
  unlink(f.c_str());
}

void EpicProc::init() 
{ 
  PROC
  init_pimpl();
  setOalInstrument();

  if(ePS_->mypars["filterevents"]->boolValue()){
    ePS_->has_filterexpression = true;
    ePS_->filter_expression =
      ePS_->mypars["filterexpression"]->stringValue();
    // negate
    ePS_->destructive_filter = !ePS_->mypars["flagfilteredevents"]->boolValue();
  }

  if(ePS_->mypars["withgtiset"]->boolValue())
    ePS_->observation_gtis.push_back(ePS_->mypars["gtiset"]->stringValue());
  
  ENDPROC 
}

void EpicProc::setOalInstrument() const
{
  oalServer.instrument(ePS_->inst.id());
}

void EpicProc::transferParameter(const std::string& par, ParameterInfo& task) const {
  task[par]->setValue(ePS_->mypars[par]);
}


void EpicProc::process()
{
  PROC
  selectExposures();
  processAttitude();
  processHk();
  typedef  std::vector<Proc::ScienceExposure>::iterator EI;
  for(EI K = ePS_->exposures.begin(); K != ePS_->exposures.end(); ++K){
    try {
      preProcessExposure(*K);
      processExposure(*K);
      postProcessExposure(*K);
    }
    catch(ProcException e){
      ePS_->proc.print(*K, errstr);
      errstr << Sas::warning("NotFullyProcessed");
    }
  }
  ENDPROC    
}

void EpicProc::selectExposures()
{
  PROC
    // If the user is using instexpids, ignore any mode selection.

  if(! ePS_->mypars["withinstexpids"]->boolValue()){
    if(ePS_->mypars["selectmodes"]->boolValue()){
      for(map<string, Xmm::DataMode>::const_iterator K =
	    ePS_->allowed_modes.begin(); K != ePS_->allowed_modes.end(); ++K)	
	if(ePS_->mypars[(*K).first]->boolValue())
	  ePS_->datamodes.push_back((*K).second);
    } else 
      ePS_->datamodes.push_back(ePS_->default_mode);
  } else { // ... all modes will be considered
    for(map<string, Xmm::DataMode>::const_iterator K =
	  ePS_->allowed_modes.begin(); K != ePS_->allowed_modes.end(); ++K)	
      ePS_->datamodes.push_back((*K).second);
  }

  for(unsigned k=0; k<ePS_->datamodes.size(); ++k)
    errstr << "Selected data mode: " << ePS_->datamodes[k]
	   << Sas::message(Msg::Meta, Msg::Silent);
  
  if(ePS_->mypars["selectccds"]->boolValue()){
    for(unsigned ccd = 1; ccd <= ePS_->number_of_ccds; ++ccd){
      stringstream par;
      par << "ccd" << ccd;
      if(ePS_->mypars[par.str()]->boolValue())
	ePS_->ccds.push_back(ccd);
    }
  } else
    for(unsigned ccd=1; ccd<= ePS_->number_of_ccds; ++ccd)
      ePS_->ccds.push_back(ccd);
  
  for(unsigned ccd=0; ccd<ePS_->ccds.size(); ++ccd)
    errstr << "Selected CCD: " << ePS_->ccds[ccd]
	   << Sas::message(Msg::Meta, Msg::Silent);
  
  vector<string> instexpids;
  
  if(ePS_->mypars["withinstexpids"]->boolValue())
    instexpids = ePS_->mypars["instexpids"]->stringListValue();
  else {
    for(unsigned k=0; k<ePS_->datamodes.size(); ++k){
      oalServer.dataMode(ePS_->datamodes[k]);
      unsigned n = oalServer.odf()->exposures();
      errstr << ePS_->datamodes[k] << " -> " << n
	     << Sas::message(Msg::Meta, Msg::Sparse);
      for(unsigned j=1; j<=n ; ++j){
	oalServer.exposureIndex(j);
	instexpids.push_back(oalServer.instrExposureId());
      }
    }
  }
  
  // Weed out multiple exposures
  sort(instexpids.begin(), instexpids.end());
  vector<string>::iterator p = unique(instexpids.begin(), instexpids.end());
  instexpids.erase(p, instexpids.end());
  //  print_vector(instexpids);

  // For MOS in timing mode setting the instrExposureId() only
  // identifies the central CCD (the one in timing mode), whereas we
  // also want the corresponding imaging exposure (for the external
  // CCDs). We break this ambiguity by looping over all modes, for
  // each mode over all exposure indexes, we ask each exposure
  // so-selected to give us its instrExposureId, and if the latter is
  // on the list we have built above, then we construct a
  // ScienceExposure.

  setOalInstrument();
  for(unsigned l = 0; l < ePS_->datamodes.size(); l++){
    oalServer.dataMode(ePS_->datamodes[l]);
    unsigned N = oalServer.odf()->exposures();

    errstr << "Considering one of the " << N << " exposures for mode "
	   << oalServer.dataMode() << Sas::message(Msg::Meta, Msg::Silent);

    for(unsigned n = 0; n < N; n++){
      oalServer.exposureIndex(n+1);

      std::string id = oalServer.instrExposureId();

      errstr << "Considering exposure " << id
	     << "[index=" << n << "]" << Sas::message(Msg::Meta, Msg::Silent);
      

      if(std::find(instexpids.begin(), instexpids.end(), id) != instexpids.end()){
	
	// Make sure the user is not asking us to process exposures from
	// some other instrument.
	if(oalServer.instrument() != ePS_->inst.id()){
	  Sas::message(Msg::Meta, Msg::Silent, oalServer.instrExposureId() +
		       " discarded because it belongs to another instrument.");
	  continue;
	}
// 	if(find(ePS_->datamodes.begin(), ePS_->datamodes.end(),
// 	    oalServer.dataMode()) == ePS_->datamodes.end()){
// 	  Sas::message(Msg::Meta, Msg::Silent, (*K) +
// 		       " discarded because its mode is not selected.");      
// 	  continue;
// 	}
	Proc::ScienceExposure s;
	s.mode = oalServer.dataMode();
	s.index = n;
	if(oalServer.odf()->hasSet(Xmm::AUXILIARY)){
	  s.auxset.set = oalServer.odf()->selectSet(Xmm::AUXILIARY);
	  s.hasauxset = true;
	} else
	  s.hasauxset = false;
	s.exposure = oalServer.odf()->currentExposure();
	ePS_->exposures.push_back(s);
      } else {
	errstr << "It is not on my list. Skipped." <<
	  Sas::message(Msg::Meta, Msg::Silent);
      }
    }
  }
  setOalInstrument(); // In case the user gave us an exposure from
                      // some other instrument.
  for(unsigned k=0; k<ePS_->exposures.size(); ++k)
    errstr << "Selected exposure: " << ePS_->exposures[k].exposure->counter()
	   << " mode " << ePS_->exposures[k].mode
	   << Sas::message(Msg::Meta, Msg::Silent);
  ENDPROC
}


void EpicProc::processAttitude()
{
  PROC

    if(dataSetServer->exists(ePS_->proc.atthkgenname())){
      Sas::message(Msg::Meta, Msg::Silent, ePS_->proc.atthkgenname() +
		   " exists. atthkgen will not be run again.");
      ePS_->has_atthkgenset = true;
    } else {
      string rp = ePS_->mypars["referencepointing"]->stringValue();
      if(ePS_->mypars["runatthkgen"]->boolValue() || rp == "mean"
	 || rp == "median"){
	Info::Task* atthkgen = task("atthkgen");
	ParameterInfo& atthkgenpars = *atthkgen->parameters();
	atthkgenpars.revertToDefault();
	atthkgenpars["atthkset"]->setValue(ePS_->proc.atthkgenname());
	if(ePS_->proc.protectedrun(atthkgen))
	  ePS_->has_atthkgenset = true;
	else {
	  ePS_->has_atthkgenset = false;
	  // make sure we don't leave an incomplete file behind
	  unlink(ePS_->proc.atthkgenname().c_str());
	}
      }
    }

 ENDPROC
}

void EpicProc::processHk()
{
  PROC
    if(dataSetServer->exists(ePS_->proc.hkgtiname())){
      Sas::message(Msg::Meta, Msg::Silent, ePS_->proc.hkgtiname() +
		   " exists. hkgtigen will not be run again.");
      ePS_->has_hkgti = true;
    } else {
      if(ePS_->mypars["runhkgtigen"]->boolValue())
	{
	  Info::Task* hkgtigen = task("hkgtigen");
	  ParameterInfo& hkgtigenpars = *hkgtigen->parameters();
	  hkgtigenpars.revertToDefault();
	  
	  
	  switch(ePS_->inst.id()){
	  case Xmm::EPN:
	    hkgtigenpars["instrument"]->setValue("epn");
	    break;
	  case Xmm::EMOS1:
	    hkgtigenpars["instrument"]->setValue("emos1");
	    break;
	  case Xmm::EMOS2:
	    hkgtigenpars["instrument"]->setValue("emos2");
	    break;
	  default:
	    errstr << "The instrument seems to be set to "
		   << ePS_->inst.id() << " and I cannot cope with this. Report it. Switch hkgtigen off. " << Sas::error("IncorrectInstrument");	    
	  }
	  	  
	  hkgtigenpars["withparameters"]->setValue(ePS_->mypars["withparameters"]);
	  hkgtigenpars["parameters"]->setValue(ePS_->mypars["parameters"]);
	  hkgtigenpars["except"]->setValue(ePS_->mypars["except"]);
	  hkgtigenpars["withoverrideparameters"]->setValue(ePS_->mypars["withoverrideparameters"]);
	  hkgtigenpars["overrideparameters"]->setValue(ePS_->mypars["overrideparameters"]);
	  hkgtigenpars["withgtiset"]->setValue(true);
	  hkgtigenpars["gtiset"]->setValue(ePS_->proc.hkgtiname());
	  if(ePS_->proc.protectedrun(hkgtigen)){
	    ePS_->observation_gtis.push_back(ePS_->proc.hkgtiname());
	    ePS_->has_hkgti = true;
	  } else {
	    ePS_->has_hkgti = false;
	    unlink(ePS_->proc.hkgtiname().c_str());
	  }
	}
    }
  ENDPROC
}

/*
  void EpicProc::processExposure(Proc::ScienceExposure & exp){
  PROC
  
  errstr << "Processing ";
  ePS_->proc.print(exp, errstr);
  errstr << Sas::message(Msg::Meta, Msg::Silent);
  
  const list<const OdfSet*>& l = exp.exposure->sets();
  typedef std::list<const OdfSet*>::const_iterator LI;
  for(LI i=l.begin(); i!=l.end(); ++i){
    const OdfSet* os = (*i);
    if(find(ePS_->ccds.begin(), ePS_->ccds.end(), os->ccdNr()) != ePS_->ccds.end()){
      try {
	frames(os);   	 
	badpixels(os);
	//AI: Include the new call to epreject
	if(ePS_->inst.id() == Xmm::EPN )
	  epreject(os);
	events(os);
	attitude(os);
	finalize(os);
	filter(os);
      }
      catch(ProcException e){	
      }
    }
  }
  ENDPROC
}
*/

void EpicProc::preProcessExposure(Proc::ScienceExposure& )
{
  PROC
    
    // These are used during and exposure loop over the CCDs
    ePS_->exposure_gtis.clear();
  ePS_->ccd_gtis.clear();
  ePS_->exposure_evlists.clear();
  ePS_->post_selection_expression = "";
  ePS_->badpix_evlists.clear();
  ePS_->inFOVarea.clear();
  ePS_->exposure_tempEvlists.clear();

  ENDPROC
}


void EpicProc::badpixels(const OdfSet* os)
{
  PROC
  ePS_->has_badpixset = false;
  if(couldRunBadpixfindOn(os) &&
     ePS_->mypars["searchforbadpixels"]->boolValue())
    {
      preBadpixelsHook(os);
      try{
	badpixfind(os);
	ePS_->has_badpixset = true;
      }
      catch(...){
	Sas::message(Msg::AppLib, Msg::Silent, "badpixfind failed");
      }
    }
  
  badpix(os);
  ENDPROC
}

void EpicProc::badpixfind(const OdfSet *)
{
  PROC
    Info::Task* badpixfind = task("badpixfind");
  ParameterInfo& badpixfindpars = *badpixfind->parameters();
  badpixfindpars.revertToDefault();
  badpixfindpars["eventset"]->setValue(ePS_->proc.badpixfindeventsname());
  badpixfindpars["badpixset"]->setValue(ePS_->proc.badpixfindname());
  
  badpixfindpars["threshabovebackground"]->setValue(ePS_->mypars["threshabovebackground"]);

  badpixfindpars["withsearchbadcolumn"]->setValue(ePS_->mypars["searchforbadcolumns"]);
  badpixfindpars["columnsearchlabel"]->setValue(ePS_->mypars["columnsearchlabel"]);
  
  badpixfindpars["thresholdlabel"]->setValue(ePS_->mypars["thresholdlabel"]);
  badpixfindpars["lothresh"]->setValue(ePS_->mypars["lothresh"]);
  badpixfindpars["hithresh"]->setValue(ePS_->mypars["hithresh"]);
  
  badpixfindpars["loenergythresh"]->setValue(ePS_->mypars["loenergythresh"]);
  badpixfindpars["hienergythresh"]->setValue(ePS_->mypars["hienergythresh"]);
  
  badpixfindpars["narrowerthanpsf"]->setValue(ePS_->mypars["narrowerthanpsf"]);
  
  badpixfindpars["backgroundrate"]->setValue(ePS_->mypars["backgroundrate"]);
  
  badpixfindpars["flickerchisqthresh"]->setValue(ePS_->mypars["flickerchisqthresh"]);
  badpixfindpars["flickerksthresh"]->setValue(ePS_->mypars["flickerksthresh"]);
  badpixfindpars["flickertimesteps"]->setValue(ePS_->mypars["flickertimesteps"]);
  
  badpixfindpars["hicolthresh"]->setValue(ePS_->mypars["hicolthresh"]);
  badpixfindpars["locolthresh"]->setValue(ePS_->mypars["locolthresh"]);    
  
  if(!ePS_->proc.protectedrun(badpixfind))
    throw(ProcException(""));

  ENDPROC
}


void EpicProc::attitude(const OdfSet*)
{
  PROC
    Info::Task* attcalc = task("attcalc");
  ParameterInfo& attcalcpars = *attcalc->parameters();
  attcalcpars.revertToDefault();
  attcalcpars["withatthkset"]->setValue(ePS_->has_atthkgenset);
  attcalcpars["atthkset"]->setValue(ePS_->proc.atthkgenname());
  attcalcpars["eventset"]->setValue(ePS_->proc.eventsname());
  string rp = ePS_->mypars["referencepointing"]->stringValue();
  if(rp == "nominal")
    attcalcpars["refpointlabel"]->setValue("nom");
  else if(rp == "object")
    attcalcpars["refpointlabel"]->setValue("obj");
  else if(rp == "mean"){
    attcalcpars["refpointlabel"]->setValue("pnt");
    attcalcpars["withmedianpnt"]->setValue(false);
  }
  else if(rp == "median"){
    attcalcpars["refpointlabel"]->setValue("pnt");
    attcalcpars["withmedianpnt"]->setValue(true);
  }
  else if(rp == "user"){
    attcalcpars["fixeddec"]->setValue(ePS_->mypars["dec"]);
    attcalcpars["fixedra"]->setValue(ePS_->mypars["ra"]);
    attcalcpars["fixedposangle"]->setValue(ePS_->mypars["posangle"]);
  }
  if(!ePS_->proc.protectedrun(attcalc))
    throw(ProcException(""));
  ENDPROC
}	


void EpicProc::filter(const OdfSet*) {
  PROC
  
  // Here we only filter with an event removing filter. GTI can be
  // done only on the combined event list.
  if(ePS_->has_filterexpression){
    string expression = ePS_->filter_expression;
    Info::Task* evselect = task("evselect");
    ParameterInfo& evselectpars = *evselect->parameters();
    evselectpars.revertToDefault();
    evselectpars["table"]->setValue(ePS_->proc.eventsname() + ":EVENTS");
    evselectpars["destruct"]->setValue(ePS_->destructive_filter);
    evselectpars["expression"]->setValue(expression);
    evselectpars["writedss"]->setValue(true);
    evselectpars["updateexposure"]->setValue(false);
    evselectpars["keepfilteroutput"]->setValue(true);
    if(!ePS_->proc.protectedrun(evselect))
      throw(ProcException(""));
  }

ENDPROC
  }

void EpicProc::finalize(const OdfSet * os)
{
  // Add the observation-level and exposure-level GTIs to the
  // list of CCD-level GTIs.
  bool enablegtiprocessing = false;
  copy(ePS_->exposure_gtis.begin(), ePS_->exposure_gtis.end(),
       back_inserter(ePS_->ccd_gtis));
  copy(ePS_->observation_gtis.begin(), ePS_->observation_gtis.end(),
       back_inserter(ePS_->ccd_gtis));
  
  if(!ePS_->ccd_gtis.empty()){
    Info::Task* gtimerge = task("gtimerge");
    ParameterInfo& gtimergepars = *gtimerge->parameters();
    gtimergepars.revertToDefault();
    gtimergepars["withgtitable"]->setValue(true);
    gtimergepars["gtitable"]->setValue(ePS_->proc.ccdgtitablename(true));
    //	  gtimergepars["gtitable"]->setValue(ePS_->proc.eventsname() + ":STDGTI");

    // AI 18/11/2004
    // Scisim needs to run gtimerge with the mode "OR" in stead of "AND"
    // In principle, if SciSim background is not enable, the AUX file does not produce
    // frames in all the CCDs, and therefore the GTI are not well defined.
   
    std::string gtimergemode;
    if ( booleanParameter("analyzingSciSimdata") )
      gtimergemode = "or";
    else
      gtimergemode = "and";
      
    gtimergepars["mergemode"]->setValue(gtimergemode);

    vector<string> tables(ePS_->ccd_gtis.size());
    for(unsigned j=0; j<ePS_->ccd_gtis.size(); ++j)
      tables[j] = ePS_->ccd_gtis[j] + ":";
    if(ePS_->has_observationlevelgti)
      tables.push_back(ePS_->proc.observationgtiname() + ":STDGTI");
    gtimergepars["tables"]->setValue(tables);
    if(ePS_->proc.protectedrun(gtimerge)){
      enablegtiprocessing = true;
    }
  }
    
  try {
    finalizeHook(os);
  }
  catch(...){
    throw(ProcException(""));
  }

  if(ePS_->mypars["applygti"]->boolValue() && enablegtiprocessing){
    stringstream s;
    s << " CCDNR == " << oalServer.ccd() << " && TIME in gti("
      << ePS_->proc.ccdgtitablename(true) << ")" ;
    // FIXME: I would like to filter on GTI here as well but I
    // cannot.
    // expression += string(" && ") + s.str();
    if(ePS_->post_selection_expression.size() > 0)
      ePS_->post_selection_expression += "||";
    else
      ePS_->post_selection_expression = "(";
    ePS_->post_selection_expression += string("(") + s.str() + string(")");
  }

}

void EpicProc::finalizeHook(const OdfSet *){ PROC ENDPROC }

void EpicProc::postProcessExposure(Proc::ScienceExposure& exp)
{


  if(ePS_->post_selection_expression.size() > 0)
    ePS_->post_selection_expression += ")";
  
  combine(ePS_->exposure_evlists, exp);

  if(ePS_->mypars["removeintermediateeventlists"]->boolValue())
    for_each(ePS_->exposure_evlists.begin(), ePS_->exposure_evlists.end(), Unlink);

  if(ePS_->mypars["removetemporaries"]->boolValue())
    ePS_->proc.remove_temporaries();

  if(ePS_->mypars["removeintermediategtis"]->boolValue())
    ePS_->proc.remove_gtis();


}


void EpicProc::combine(const vector<string>& el,
		       const Proc::ScienceExposure& thisexp)
{
  PROC
    string eventlist;
  bool removeeventlists = ePS_->mypars["removeintermediateeventlists"]->boolValue();
  if(ePS_->mypars["runevlistcomb"]->boolValue()){
    if(el.size() == 0){
      ePS_->proc.print(thisexp, errstr);
      errstr << Sas::warning("NoEventListsToMerge");
    } else {
      Info::Task* evlistcomb = task("evlistcomb");
      ParameterInfo& evlistcombpars = *evlistcomb->parameters();      
      evlistcombpars.revertToDefault();

      /*
       * SPR-SOC-2625
       * evlistcomb has to be called with the correct 
       * instrument name. If not, the final event list
       * does not contain the OFFSETY column
       */
      switch(ePS_->inst.id())
	{
	case Xmm::EPN:
	  evlistcombpars["instrument"]->setValue("epn");
	  break;
	case Xmm::EMOS1: 
	case Xmm::EMOS2:
	  evlistcombpars["instrument"]->setValue("emos");
	  break;
	default:
	  errstr << "The instrument seems to be set to "
		 << ePS_->inst.id() << " and I cannot cope with this. Report it. Switch hkgtigen off. " 
		 << Sas::error("IncorrectInstrument");	    
	}

      /*
       * SPR-SOC-2625
       */


      evlistcombpars["eventsets"]->setValue(el);
      
      // FIXME: the user of evlistcomb is _not_ supposed to change the
      // value of the parameter "e(mos|pn)datamodes". See SSC-SPR-2392.
      if(thisexp.mode == Xmm::IMAGING ||
	 thisexp.mode == Xmm::REDUCED_IMAGING){	
	eventlist = ePS_->proc.imagingeventsname();
      } else if(thisexp.mode == Xmm::TIMING ||
		thisexp.mode == Xmm::COMPRESSED_TIMING ||
		thisexp.mode == Xmm::BURST){
	eventlist = ePS_->proc.timingeventsname();
      }
      
      // Set them both, but only one will be used by evlistcomb
      evlistcombpars["imagingset"]->setValue(eventlist);
      evlistcombpars["timingset"]->setValue(eventlist);
      
      // epproc should run evlistcomb with the parameter
      // OFFSET in the maintable option. See SOC-SPR-2538
      vector<string> maintbl;
      maintbl.push_back("EVENTS");
      maintbl.push_back("OFFSETS");
      evlistcombpars["maintable"]->setValue(maintbl);
      //See SOC-SPR-2538
      
      combineHook(evlistcombpars);
      
      if(ePS_->proc.protectedrun(evlistcomb)){
	Info::Task* dscopyblock = task("dscopyblock");
	ParameterInfo& dscopyblockpars = *dscopyblock->parameters();
	dscopyblockpars.revertToDefault();
	vector<string> blks(1);
	blks[0] = calServer.ccf()->cifName() + ":CALINDEX";
	dscopyblockpars["to"]->setValue(eventlist);
	dscopyblockpars["blocks"]->setValue(blks);
	ePS_->proc.protectedrun(dscopyblock);
	if(ePS_->has_filterexpression){
	  Info::Task* evselect = task("evselect");
	  ParameterInfo& evselectpars = *evselect->parameters();
	  evselectpars.revertToDefault();
	  evselectpars["table"]->setValue(eventlist + ":EVENTS");
	  evselectpars["destruct"]->setValue(ePS_->destructive_filter);
	  evselectpars["expression"]->setValue(ePS_->post_selection_expression);
	  evselectpars["writedss"]->setValue(true);
	  evselectpars["updateexposure"]->setValue(true);
	  evselectpars["keepfilteroutput"]->setValue(true);
	  if(!ePS_->proc.protectedrun(evselect))
	    Sas::message(Msg::Meta, Msg::Silent, "Post evlistcomb filtering on " + eventlist + " failed.");
	  
	  /*
	   * screen EXPOSUxx extensions: delete non-standard columns
	   */
	  screenExposure(thisexp);
	  
	  
	}
      } else {
	if(removeeventlists){
	  // Remove the intermediate event lists only if the combination
	  // stage was successful.
	  removeeventlists = false;
	  errstr << "The event list combination stage failed for ";
	  ePS_->proc.print(thisexp, errstr);
	  errstr << ". The intermediate event lists will not be removed."
		 << Sas::warning("IntermediateEventListsNotRemoved");
	}
      }
    }
  }
  if(removeeventlists)
    for(unsigned k = 0; k < el.size(); ++k)
      unlink(el[k].c_str());
  
  ENDPROC
    }


EpicProc::~EpicProc() {
  PROC
    //  cleanup();
    if(ePS_->mypars["removeintermediategtis"]->boolValue())
    ePS_->proc.remove_last();
  ENDPROC
}

void EpicProc::preProcessExposureHook(Proc::ScienceExposure&) {}
void EpicProc::postProcessExposureHook(Proc::ScienceExposure&) {}
void EpicProc::preBadpixelsHook(const OdfSet *) {}
void EpicProc::screenExposure(const Proc::ScienceExposure&){ PROC ENDPROC }
