//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// October 2000
//
#include <Emproc.h>

#include <Dal.h>	// SAS
#include <OalServer.h>
#include <CalServer.h>
#include <param.h>
#include <errstr.h>
#include <metatask.h>
#include <Evaluator.h>

#include <badpix.h>
#include <emframes.h>
#include <emevents.h>
#include <emenergy.h>

//AI: Headers for new task
#include <emeventsproj.h>
#include <embadpixfind.h>
#include <ebadpixupdate.h>
#include <tabgtigen.h>


#include <algorithm>
#include <cstdlib>
#include <unistd.h>
#include <iomanip>

Emproc::Emproc(Xmm::Instrument ins)
{
  PROC

  ePS_ = new EpicProcState(7,
			   xmmPayLoad.instrument(ins),
			   Xmm::IMAGING);  
  init();

  ENDPROC
}  

void Emproc::init_pimpl(){
  PROC
  
    // Move to constructor as we know there which of the MOSes we are
    // supposed to be. 
    //    ePS_ = new EpicProcState(7,
    //  			   xmmPayLoad.instrument(Xmm::EMOS1),
    //  			   Xmm::IMAGING);
  
  ePS_->allowed_modes.insert(make_pair(string("imaging"), Xmm::IMAGING));
  ePS_->allowed_modes.insert(make_pair(string("rimaging"), Xmm::REDUCED_IMAGING));
  ePS_->allowed_modes.insert(make_pair(string("timing"), Xmm::TIMING));
  ePS_->allowed_modes.insert(make_pair(string("ctiming"), Xmm::COMPRESSED_TIMING));
  
  ENDPROC
}


void Emproc::processExposure(Proc::ScienceExposure & exp){
  PROC

  // Define criterion for GATTI flare screening
  // Corresponds to area in arcmin (Small Window is 3.7)
  double areaCcdMin = 3.0;
  double inFOVarea = 0;

  errstr << "Processing ";
  ePS_->proc.print(exp, errstr);
  errstr << Sas::message(Msg::Meta, Msg::Silent);
  
  const list<const OdfSet*>& l = exp.exposure->sets();
  typedef std::list<const OdfSet*>::const_iterator LI;
  for(LI i=l.begin(); i!=l.end(); ++i)
    {
      const OdfSet* os = (*i);
      if(find(ePS_->ccds.begin(), ePS_->ccds.end(), os->ccdNr()) != ePS_->ccds.end())
	{
	  try 
	    {
	      frames(os);   	
	      if(couldRunBadpixfindOn(os) &&
		 ePS_->mypars["searchforbadpixels"]->boolValue())
		preBadpixelsHook(os);  
	      badPixelsSearch(os);
	    }
	  catch(ProcException e) { }
	}
    }

  
 
 
  if (ePS_->mypars["searchforbadpixels"]->boolValue())
    {
      if(ePS_->mypars["badpixfindalgo"]->stringValue() == "embadpixfind")
	{
	  /*
	   * Set area Ccd value.
	   */ 
	  for (unsigned int i = 0 ; i < ePS_->inFOVarea.size() ; ++i)
	    inFOVarea += ePS_->inFOVarea[i];
	  
	  ePS_->areaCcd = inFOVarea;         
	  
	  if (inFOVarea > areaCcdMin )
	    {
	      
	      //Call to evlistcomb, evselect and ebadpixupdate
	      mergeIntermediateResults(exp);
	      badPixUpdate();
	      /*
	       * makeFlareTs filters the combined event list searching for 
	       * single events within field of view flagged by Gatti. It is possible 
	       * (but weird) that there were not filtered events at all. Therefore, it is 
	       * possible to obtain an empty file, in that case, the GATTI flare screening 
	       * has no sense and we skip to findlowener branch.
	       */
	      if (makeFlareTs())
		{
		  tsToGTI();
		  /*
		   * Test whether flare GTI should be applied
		   */
		  testFlareGTI();		 
		}
	      else 
		/*
		 * Disable the second iteration for embadpixfind
		 */ 
		ePS_->hasValidGTI = false;		
	    }
	  else
	    {
	      /*
	       * Disable the second iteration for embadpixfind
	       */ 
	      ePS_->hasValidGTI = false;
	    }	    
	}
    }
  else 
    {
      /*
       * Disable the second iteration for embadpixfind
       */ 
      ePS_->hasValidGTI = false;
    }
  
  

  for(LI j=l.begin(); j!=l.end(); ++j)
    {
      const OdfSet* os = (*j);
      if(find(ePS_->ccds.begin(), ePS_->ccds.end(), os->ccdNr()) != ePS_->ccds.end())
	{
	  try 
	    {
	      if(couldRunBadpixfindOn(os) &&
		 ePS_->mypars["searchforbadpixels"]->boolValue())		
		findFinalBadPixels(os);

	      badpix(os);       	   
	      events(os);
	      attitude(os);
	      finalize(os); //?
	      filter(os);
	    }
	  catch(ProcException e) { }
	}
    }

  ENDPROC
}


void Emproc::frames(const OdfSet* os)
{
  PROC

  errstr << "Processing data set: " << os->name()
	 << Sas::message(Msg::Meta, Msg::Silent);

  oalServer << os->name();


  Info::Task* emframes = task("emframes");
  ParameterInfo& emframespars = *emframes->parameters();    
  emframespars.revertToDefault();
  if(oalServer.odf()->hasSet(Xmm::AUXILIARY))
    emframespars["auxiliaryset"]->setValue(oalServer.odf()->selectSet(Xmm::AUXILIARY)->name());
  else
    throw(ProcException(""));  
  emframespars["frameset"]->setValue(ePS_->proc.framesname());
  emframespars["withodfeventset"]->setValue(true);
  emframespars["odfeventset"]->setValue(os->name());
  emframespars["neweventset"]->setValue(true);
  emframespars["outeventset"]->setValue(ePS_->proc.frameseventsname());
  emframespars["writegtiset"]->setValue(true);
  emframespars["outgtiset"]->setValue(ePS_->proc.framesgtiname());
  emframespars["flagbadtimes"]->setValue(ePS_->has_hkgti);
  emframespars["ingtiset"]->setValue(ePS_->proc.hkgtiname());
  if(os->dataMode() == Xmm::TIMING ||
     os->dataMode() == Xmm::COMPRESSED_TIMING){
    emframespars["withsrccoords"]->setValue(ePS_->mypars["withsrccoords"]);
    emframespars["srcra"]->setValue(ePS_->mypars["srcra"]);
    emframespars["srcdec"]->setValue(ePS_->mypars["srcdec"]);
  }
  
      
  
  if(!ePS_->proc.protectedrun(emframes))
    throw(ProcException(""));
  
  ePS_->ccd_gtis.push_back(ePS_->proc.framesgtiname());

  /*
   * AI: CHECK ID IT NEEDED!!!!! IF NOT REMOVE!!
   * AND ALSO REMOVE merge_selection_expression FROM Proc.cc
  // Build the expression to use in mergeIntermediateResults
  if(ePS_->mypars["applygti"]->boolValue()){
  stringstream s;
  s << "CCDNR == " << oalServer.ccd() << " && TIME in gti("
  << ePS_->proc.ccdgtitablename(true) << ")" ;
  
  if(ePS_->merge_selection_expression.size() > 0)
  ePS_->merge_selection_expression += "||";
  else
  ePS_->merge_selection_expression = "(";
  ePS_->merge_selection_expression += string("(") + s.str() + string(")");
  }
  */

  ENDPROC
}

bool Emproc::couldRunBadpixfindOn(const OdfSet * os)
{
  PROC
    RETURN (os->dataMode() == Xmm::IMAGING ||
	    os->dataMode() == Xmm::REDUCED_IMAGING);
}

void Emproc::preBadpixelsHook(const OdfSet * os)
{
  PROC

  // Identify the offset/variance sets corresponding to this CCD
  // emevents decides which of these are worth using.
  vector<string> OVEsets;
  oalServer.dataMode(Xmm::DATA_OFFSET_VARIANCE);
  for(unsigned j=1; j<=oalServer.odf()->exposures(); ++j){
    oalServer.exposureIndex(j);	
    if(oalServer.odf()->hasSet())
      if(oalServer.odf()->selectSet()->ccdNr() == os->ccdNr())
	OVEsets.push_back(oalServer.odf()->selectSet()->name());
  }
  if(!OVEsets.empty()){
    errstr << "Found the following offset/variance data sets: "<< endl;
    for(unsigned k=0; k<OVEsets.size(); ++k)
      errstr << OVEsets[k] << endl;
    errstr << Sas::message(Msg::Meta, Msg::Verbose);
  }
  oalServer << os->name(); // restore state to process current data set

  Info::Task* emevents = task("emevents");
  ParameterInfo& emeventspars = *emevents->parameters();
  emeventspars.revertToDefault();

  emeventspars["odfeventset"]->setValue(ePS_->proc.frameseventsname());
  emeventspars["eventset"]->setValue(ePS_->proc.badpixfindeventsname());
  emeventspars["withframeset"]->setValue(true);
  emeventspars["frameset"]->setValue(ePS_->proc.framesname());

  emeventspars["withoffvarsets"]->setValue(!OVEsets.empty());
  emeventspars["offvarsets"]->setValue(OVEsets);

  emeventspars["analysepatterns"]->setValue(false);
  emeventspars["flagbadpixels"]->setValue(false);
  emeventspars["splitdiagonals"]->setValue(false);
  emeventspars["randomizeposition"]->setValue(false);

  if(os->dataMode() != Xmm::IMAGING)
    emeventspars["rejectbade3"]->setValue(false);	  
  else
    emeventspars["rejectbade3"]->setValue(true);


  if(!ePS_->proc.protectedrun(emevents)){
    throw(ProcException(""));
  }

  if(os->dataMode() == Xmm::IMAGING){
    Info::Task* emenergy = task("emenergy");
    ParameterInfo& emenergypars = *emenergy->parameters();
    emenergypars.revertToDefault();
    emenergypars["ineventset"]->setValue(ePS_->proc.badpixfindeventsname());
  
    emenergypars["correctcti"]->setValue(false);
    emenergypars["correctgain"]->setValue(false);	      
    emenergypars["randomizeenergy"]->setValue(false);
    
    if(!ePS_->proc.protectedrun(emenergy))
      throw(ProcException(""));
  }
  
  std::string expression("(FLAG & 0x762aa000) == 0");
  if(ePS_->mypars["withbadpixgti"]->boolValue())
    expression += "&& TIME in GTI(" +
      ePS_->mypars["badpixgti"]->stringValue() + ")";
  
  Info::Task* evselect = task("evselect");
  ParameterInfo& evselectpars = *evselect->parameters();
  evselectpars.revertToDefault();
  evselectpars["table"]->setValue(ePS_->proc.badpixfindeventsname() + ":EVENTS");
  evselectpars["keepfilteroutput"]->setValue(true);
  evselectpars["destruct"]->setValue(true);
  evselectpars["keepfilteroutput"]->setValue(true);
  evselectpars["writedss"]->setValue(true);
  evselectpars["updateexposure"]->setValue(false);
  evselectpars["expression"]->setValue(expression);
  
  if(!ePS_->proc.protectedrun(evselect))
      throw(ProcException(""));

  ENDPROC
}


void Emproc::badpix(const OdfSet* os)
{
  PROC

  Info::Task* badpix = task("badpix");
  ParameterInfo& badpixpars = *badpix->parameters();
  badpixpars.revertToDefault();
  badpixpars["badpixset"]->setValue(ePS_->proc.badpixfindname());
  badpixpars["eventset"]->setValue(ePS_->proc.frameseventsname());
  if(os->dataMode() == Xmm::TIMING ||
     os->dataMode() == Xmm::COMPRESSED_TIMING){
    badpixpars["getuplnkbadpix"]->setValue(true);
    badpixpars["getotherbadpix"]->setValue(false);
    badpixpars["getnewbadpix"]->setValue(false);
    badpixpars["emptyextension"]->setValue(false);
  } else {
    badpixpars["getuplnkbadpix"]->setValue(true);
    badpixpars["getotherbadpix"]->setValue(true);
    badpixpars["getnewbadpix"]->setValue(ePS_->has_badpixset);
  }
  badpixpars["windowfilter"]->setValue(true);

  if(!ePS_->proc.protectedrun(badpix))
    throw(ProcException(""));
  
  ENDPROC
}

void Emproc::events(const OdfSet* os)
{
  PROC

    // Identify the offset/variance sets corresponding to this CCD
    // emevents decides which of these are worth using.
    vector<string> OVEsets;
  oalServer.dataMode(Xmm::DATA_OFFSET_VARIANCE);
  for(unsigned j=1; j<=oalServer.odf()->exposures(); ++j){
    oalServer.exposureIndex(j);	
    if(oalServer.odf()->hasSet())
      if(oalServer.odf()->selectSet()->ccdNr() == os->ccdNr())
	OVEsets.push_back(oalServer.odf()->selectSet()->name());
  }
  if(!OVEsets.empty()){
    errstr << "Found the following offset/variance data sets: "<< endl;
    for(unsigned k=0; k<OVEsets.size(); ++k)
      errstr << OVEsets[k] << endl;
    errstr << Sas::message(Msg::Meta, Msg::Verbose);
  }
  oalServer << os->name(); // restore state to process current data set

  Info::Task* emevents = task("emevents");
  ParameterInfo& emeventspars = *emevents->parameters();
  emeventspars.revertToDefault();

  emeventspars["odfeventset"]->setValue(ePS_->proc.frameseventsname());
  emeventspars["eventset"]->setValue(ePS_->proc.eventsname());
  emeventspars["withframeset"]->setValue(true);
  emeventspars["frameset"]->setValue(ePS_->proc.framesname());

  emeventspars["withoffvarsets"]->setValue(!OVEsets.empty());
  emeventspars["offvarsets"]->setValue(OVEsets);

  if(os->dataMode() != Xmm::IMAGING)
    emeventspars["rejectbade3"]->setValue(false);	  
  else {
    emeventspars["rejectbade3"]->setValue(true);
  }

  if(os->dataMode() == Xmm::COMPRESSED_TIMING)
    emeventspars["flagtruncatede1"]->setValue(false);    

  if(os->dataMode() == Xmm::TIMING ||
     os->dataMode() == Xmm::COMPRESSED_TIMING)
    emeventspars["splitdiagonals"]->setValue(false);

  if(os->dataMode() == Xmm::COMPRESSED_TIMING){
    emeventspars["analysepatterns"]->setValue(false);
    emeventspars["flagbadpixels"]->setValue(false);
  }

  if(os->dataMode() == Xmm::IMAGING ||
     os->dataMode() == Xmm::REDUCED_IMAGING)
    emeventspars["randomizetime"]->setValue(ePS_->mypars["randomizetime"]);
  
  emeventspars["randomizeposition"]->setValue(ePS_->mypars["randomizeposition"]);
  emeventspars["keepsifluor"]->setValue(ePS_->mypars["keepsifluor"]);
  emeventspars["rejectrows"]->setValue(ePS_->mypars["rejectrows"]);
  emeventspars["maxeventsperrow"]->setValue(ePS_->mypars["maxeventsperrow"]);
    
  if(!ePS_->proc.protectedrun(emevents)){
    throw(ProcException(""));
  }


  Info::Task* emenergy = task("emenergy");
  ParameterInfo& emenergypars = *emenergy->parameters();
  emenergypars.revertToDefault();
  emenergypars["ineventset"]->setValue(ePS_->proc.eventsname());

  if(os->dataMode() == Xmm::IMAGING){
    emenergypars["getccdbkg"]->setValue(true);
    emenergypars["rejectbade3e4"]->setValue(true);
  } else {
    emenergypars["getccdbkg"]->setValue(false);
    emenergypars["rejectbade3e4"]->setValue(false);
  }    
      
  emenergypars["randomizeenergy"]->setValue(ePS_->mypars["randomizeenergy"]);
  emenergypars["useccfdarkframe"]->setValue(ePS_->mypars["useccfdarkframe"]);
  //SPR-2623
  emenergypars["correctcti"]->setValue(ePS_->mypars["correctcti"]);
  emenergypars["correctgain"]->setValue(ePS_->mypars["correctgain"]);
  //SPR=2623
    
  if(!ePS_->proc.protectedrun(emenergy))
    throw(ProcException(""));


  // FIXME:
  // evselect destruct=Y keepfilteroutput=Y updateexposure=N expression="(FLAG & 0x762a0000) == 0"


  ePS_->exposure_evlists.push_back(ePS_->proc.eventsname());
  
  ENDPROC
}


void Emproc::badPixelsSearch(const OdfSet* os) 
{ 
  PROC

    // Do not run badpix unless embadpixfind or badpixfind will be executed
    ePS_->has_badpixset = false;

  if(couldRunBadpixfindOn(os) &&
     ePS_->mypars["searchforbadpixels"]->boolValue())
    {
      if(ePS_->mypars["badpixfindalgo"]->stringValue() == "embadpixfind")
	{      
	  /*
	   * emeventsproj task call
	   */
	  Info::Task* emeventsproj = task("emeventsproj");
	  ParameterInfo& emeventsprojpar = *emeventsproj->parameters();
	  emeventsprojpar.revertToDefault();
	  
	  emeventsprojpar["eventset"]->setValue(ePS_->proc.badpixfindeventsname());
	  emeventsprojpar["rejectbadevents"]->setValue(true);
	  emeventsprojpar["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	  
	  if(!ePS_->proc.protectedrun(emeventsproj))
	    throw(ProcException("")); 
	  
	  /*
	   * embadpixfind task call
	   */
	  Info::Task* embadpixfind = task("embadpixfind");
	  ParameterInfo& embadpixfindpar = *embadpixfind->parameters();
	  embadpixfindpar.revertToDefault();
	  
	  embadpixfindpar["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	  embadpixfindpar["includedeadpixels"]->setValue(true);
	  embadpixfindpar["badpixset"]->setValue(ePS_->proc.badpixfindname());
	  
	  if(!ePS_->proc.protectedrun(embadpixfind))
	    throw(ProcException("")); 
	  
	  ePS_->exposure_tempEvlists.push_back(ePS_->proc.badpixfindeventsname());
	  ePS_->badpix_evlists.push_back(ePS_->proc.badpixfindname());

	  /*
	   * areaCcdFiles:  Return area within FOV of CCD-specific event files
	   *  where GATTI flagging worked
	   */
	  DataSet*  tempDS = dataSetServer->open(ePS_->proc.badpixfindeventsname(), Dal::Read); 
	  Table* tempTab = tempDS->table("EVENTS");
	  
	  if ( (tempTab->attribute("CCDMODE")->asString() != "FAST") )
	    if (tempTab->hasAttribute("XMMEA_22") &&
		(tempTab->attribute("XMMEA_22")->asString().size() != 0) )
	      ePS_->inFOVarea.push_back(tempTab->attribute("IN_FOV")->asReal());
	  else
	    errstr << "No IN_FOV keyword in "<< ePS_->proc.badpixfindeventsname()
		   << Sas::message(Msg::App, Msg::Verbose);

	  dataSetServer->close(tempDS);	  
	  
	  //AI: Same as the call to badpixfind
	  ePS_->has_badpixset = true;
	}
      else
	{
	  /*
	   * Call to badpixfind
	   */
	  try {
	    badpixfind(os);
	    ePS_->has_badpixset = true;
	  }
	  catch(...) {
	    Sas::message(Msg::AppLib, Msg::Silent, "badpixfind failed");	    
	  }
	}      
    }

  ENDPROC
}

void Emproc::mergeIntermediateResults(Proc::ScienceExposure& thisexp )
{
  PROC  
    /*
     * Call to evlistcomb
     */
    Info::Task* evlistcomb = task("evlistcomb");
    ParameterInfo& evlistcombpars = *evlistcomb->parameters();      
    evlistcombpars.revertToDefault();
    
    evlistcombpars["eventsets"]->setValue(ePS_->exposure_tempEvlists);
 
    std::string eventlist;

    if(thisexp.mode == Xmm::IMAGING ||
       thisexp.mode == Xmm::REDUCED_IMAGING){	
      eventlist = ePS_->proc.tempImagingname();
    } else if(thisexp.mode == Xmm::TIMING ||
	      thisexp.mode == Xmm::COMPRESSED_TIMING ||
	      thisexp.mode == Xmm::BURST){
      eventlist = ePS_->proc.tempTimingname();
    }

    evlistcombpars["imagingset"]->setValue(eventlist);
    
    evlistcombpars["instrument"]->setValue("emos");
    vector<string> maintbl;
    maintbl.push_back("EVENTS");
    maintbl.push_back("OFFSETS");
    evlistcombpars["maintable"]->setValue(maintbl);
    
    combineMergeEvents(evlistcombpars);

    if(!ePS_->proc.protectedrun(evlistcomb))
      throw(ProcException(""));

    /*
     * Call to evselect:
     * Depending on the mode (Imaging, timing...)
     * we will select good events in each mode.
     */
    Info::Task* evselect = task("evselect");
    ParameterInfo& evselectpar = *evselect->parameters();
    evselectpar.revertToDefault();
    
    //Common options
    evselectpar["withfilteredset"]->setValue(true);
    evselectpar["keepfilteroutput"]->setValue(true);
    evselectpar["destruct"]->setValue(true);
    evselectpar["writedss"]->setValue(true);
    evselectpar["updateexposure"]->setValue(true);
    evselectpar["cleandss"]->setValue(true);
    
    //AI: Remove when finihed
    //ePS_->merge_selection_expression += " && ((FLAG & 0x762aa000) == 0)";
    //evselectpar["expression"]->setValue(ePS_->merge_selection_expression);

    evselectpar["expression"]->setValue("((FLAG & 0x762aa000) == 0)");


    if(thisexp.mode == Xmm::IMAGING ||
       thisexp.mode == Xmm::REDUCED_IMAGING)
      {	
	evselectpar["table"]->setValue(ePS_->proc.tempImagingname());    
	evselectpar["filteredset"]->setValue(ePS_->proc.tempImagingEvlistname());	
      } 
    else if (thisexp.mode == Xmm::TIMING ||
	     thisexp.mode == Xmm::COMPRESSED_TIMING ||
	     thisexp.mode == Xmm::BURST)
      {
	evselectpar["table"]->setValue(ePS_->proc.tempTimingname());    
	evselectpar["filteredset"]->setValue(ePS_->proc.tempTimingEvlistname());	
	
      }

    if(!ePS_->proc.protectedrun(evselect))
      throw(ProcException(""));
    
  ENDPROC
}

void Emproc::badPixUpdate()
{
  PROC

  /*
   * Call to ebadpixupdate
   */
  Info::Task* ebadpixupdate = task("ebadpixupdate");
  ParameterInfo& ebadpixupdatepar = *ebadpixupdate->parameters();
  ebadpixupdatepar["eventset"]->setValue(ePS_->proc.tempImagingEvlistname());
  ebadpixupdatepar["fromccf"]->setValue(false);
  ebadpixupdatepar["badpixtables"]->setValue(ePS_->badpix_evlists);

  if(!ePS_->proc.protectedrun(ebadpixupdate))
    throw(ProcException(""));

  ENDPROC

}

void Emproc::findFinalBadPixels(const OdfSet* os)
{
  PROC

  oalServer << os->name();

  std::string eventList;

  if (ePS_->hasValidGTI)
    {      
      /*
       * Call to evselect
       */
      Info::Task* evselect = task("evselect");
      ParameterInfo& evselectpar = *evselect->parameters();
      evselectpar.revertToDefault();
      
      evselectpar["table"]->setValue(ePS_->proc.badpixfindeventsname());
      evselectpar["withfilteredset"]->setValue(true);
      evselectpar["keepfilteroutput"]->setValue(true);
      evselectpar["updateexposure"]->setValue(false);
      
      std::string gtiExpression = "'GTI("+ePS_->proc.tempFlareBkgGTIname()+",TIME)'";
      evselectpar["expression"]->setValue(gtiExpression);
      
      evselectpar["filteredset"]->setValue(ePS_->proc.tempImagingEvlistname());
      
      /*
       * Set the event list name for the low energy search
       */
      eventList = ePS_->proc.tempImagingEvlistname();
      
      if(!ePS_->proc.protectedrun(evselect))
	throw(ProcException(""));
    }
  else
    {
      // Change the names of the file for the findlowener branch
      eventList = ePS_->proc.badpixfindeventsname();
    }
  
  
  // Do not run badpix unless embadpixfind or badpixfind will be executed
  ePS_->has_badpixset = false;
  
  if(couldRunBadpixfindOn(os) &&
     ePS_->mypars["searchforbadpixels"]->boolValue())
    {
      /*
       * Call to embadpixfind procedure
       */
      if(ePS_->mypars["badpixfindalgo"]->stringValue() == "embadpixfind")
	{
	  if (ePS_->hasValidGTI)
	    {
	      /*
	       * emeventsproj task call
	       */
	      Info::Task* emeventsproj = task("emeventsproj");
	      ParameterInfo& emeventsprojpar = *emeventsproj->parameters();
	      emeventsprojpar.revertToDefault();
	      
	      emeventsprojpar["eventset"]->setValue(ePS_->proc.tempImagingEvlistname());
	      emeventsprojpar["rejectbadevents"]->setValue(true);
	      emeventsprojpar["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	      
	      if(!ePS_->proc.protectedrun(emeventsproj))
		throw(ProcException("")); 
	      
	      /*
	       * embadpixfind task call
	       */
	      Info::Task* embadpixfind = task("embadpixfind");
	      ParameterInfo& embadpixfindpar = *embadpixfind->parameters();
	      embadpixfindpar.revertToDefault();
	      
	      embadpixfindpar["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	      embadpixfindpar["incremental"]->setValue(true);
	      embadpixfindpar["finddead"]->setValue(false);
	      embadpixfindpar["includedeadpixels"]->setValue(false);
	      embadpixfindpar["badpixset"]->setValue(ePS_->proc.badpixfindname());
	      
	      if(!ePS_->proc.protectedrun(embadpixfind))
		throw(ProcException("")); 
	      
	    }

	  /*
	   * Additional search below 500 eV
	   */
	  
	  if (ePS_->mypars["findlowener"])
	    {
	      /*
	       * Call to evselect
	       */
	      Info::Task* evselectLowEner = task("evselect");
	      ParameterInfo& evselectparLowEner = *evselectLowEner->parameters();
	      evselectparLowEner.revertToDefault();
	      
	      evselectparLowEner["table"]->setValue(eventList);
	      evselectparLowEner["withfilteredset"]->setValue(false);
	      evselectparLowEner["keepfilteroutput"]->setValue(true);
	      evselectparLowEner["expression"]->setValue("PHA < 150");
	      
	      if(!ePS_->proc.protectedrun(evselectLowEner))
		throw(ProcException(""));
	      
	      /*
	       * emeventsproj task call
	       */
	      Info::Task* emeventsprojLowEner = task("emeventsproj");
	      ParameterInfo& emeventsprojparLowEner = *emeventsprojLowEner->parameters();
	      emeventsprojparLowEner.revertToDefault();
	      
	      emeventsprojparLowEner["eventset"]->setValue(eventList);
	      emeventsprojparLowEner["rejectbadevents"]->setValue(true);
	      emeventsprojparLowEner["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	      
	      if(!ePS_->proc.protectedrun(emeventsprojLowEner))
		throw(ProcException("")); 
	  
	      /*
	       * embadpixfind task call
	       */
	      Info::Task* embadpixfindLowEner = task("embadpixfind");
	      ParameterInfo& embadpixfindparLowEner = *embadpixfindLowEner->parameters();
	      embadpixfindparLowEner.revertToDefault();
	      
	      embadpixfindparLowEner["evimageset"]->setValue(ePS_->proc.emeventsprojeventsname());
	      embadpixfindparLowEner["incremental"]->setValue(true);
	      embadpixfindparLowEner["finddead"]->setValue(true);
	      embadpixfindparLowEner["includedeadpixels"]->setValue(false);
	      embadpixfindparLowEner["badpixset"]->setValue(ePS_->proc.badpixfindname());
	      
	      if(!ePS_->proc.protectedrun(embadpixfindLowEner))
		throw(ProcException("")); 
	      
	      //AI: Same as the call to badpixfind
	      ePS_->has_badpixset = true;

	    }
	}
    }
      


  ENDPROC

}

bool Emproc::makeFlareTs()
{
  PROC

    /*
     * areaCcd : Return area within FOV of all CCDs with GATTI flagging on
     * in merged imaging file (from EXPOSUnn extensions)
     * Input: Merged events file name
     */

    DataSet*  tempDS = dataSetServer->open(ePS_->proc.tempImagingEvlistname(), Dal::Modify); 
  Table* tempTab;
  
  double areaCcd = 0;
  
  for (unsigned int i = 1 ; i<8 ; ++i)
    {
      stringstream sstr;
      sstr << "EXPOSU" << setw(2) << setfill('0')
	   << i;

      if ( tempDS->hasBlock(sstr.str()) )
	{
	  tempTab = tempDS->table(sstr.str());
	  
	  if (tempTab->hasAttribute("XMMEA_22") )
	    if (tempTab->hasAttribute("IN_FOV") )
	      areaCcd += tempTab->attribute("IN_FOV")->asReal();
	    else
	      errstr << "No IN_FOV keyword in "<< sstr.str() << " .Regenerate." 
		     << Sas::message(Msg::App, Msg::Verbose);
	  else
	    errstr << "No XMMEA_22 keyword in "<< sstr.str() << " Not found." 
		   << Sas::message(Msg::App, Msg::Verbose);
	}
    }

    /*
     * Check that the input event list contains more than 1 event
     */
  
  if( tempDS->table("EVENTS")->rows() < 2)
    {
      errstr <<ePS_->proc.tempImagingEvlistname()   
	     << " contains less than 2 events. " 
	     << "Cannot build flare screening time series." 
	     << Sas::warning("GTI flare rejection");
    }
    
  
  dataSetServer->close(tempDS);

  int areaCcdMin = 3;

  if (areaCcd < areaCcdMin)
    {
      errstr << "No CCD files with GATTI flagging. "
	     << "Flare detection not performed."
	     << Sas::warning("GTI flare rejection");
      
      return false;
    }
  else
    {

      /*
       * Call to evselect
       */
      Info::Task* evselect = task("evselect");
      ParameterInfo& evselectpar = *evselect->parameters();
      evselectpar.revertToDefault();
      
      evselectpar["table"]->setValue(ePS_->proc.tempImagingEvlistname());
      evselectpar["maketimecolumn"]->setValue(true);
      evselectpar["makeratecolumn"]->setValue(true);
      evselectpar["withrateset"]->setValue(true);
      evselectpar["updateexposure"]->setValue(false);
      evselectpar["rateset"]->setValue(ePS_->proc.tempFlareBkgTimeseriename());
      
      /*
       *  Test whether the combination of time bin and area within FOV
       *  where GATTI flagging worked is enough for flare screening
       *  Abort if too small, increase time bin if moderately small
       *
       * timebin has been definied as a parameter, although is not visible for
       * the user. Maybe we should have to remove this parameter for there and
       * put it in the ePS structure as a hard coded value.
       */
      
      double timebin;
      
      if (areaCcd*ePS_->mypars["timebin"]->realValue() < 10000)
	{
	  timebin = 10000.0/areaCcd;
	  errstr << "Time bin too small for CCD area ("<< areaCcd << " arcmin2)"
		 << " Increase to " << timebin << " for flare screening." 
		 << Sas::warning("GTI flare rejection");
	}
      else
	timebin = ePS_->mypars["timebin"]->realValue();
      
      evselectpar["timebinsize"]->setValue(timebin);
      
      
      //Keep only single events within field of view flagged by Gatti
      evselectpar["expression"]->setValue("(PATTERN==0) && ((FLAG & 0x762b8000) == 0) && #XMMEA_22");
      
      if(!ePS_->proc.protectedrun(evselect))
	throw(ProcException(""));

      /*
       * Check that the filtered event list is not empty. If it empty, we consider that
       * the GATTI rejection cannot be done and skip to findlowener branch.
       */

      DataSet*  tempDS = dataSetServer->open(ePS_->proc.tempFlareBkgTimeseriename(), Dal::Read);       
      int nrows = tempDS->table("RATE")->rows();
      dataSetServer->close(tempDS);      

      if (nrows == 0)
	{
	  errstr << "Filtered event file " << ePS_->proc.tempFlareBkgTimeseriename() 
		 << " empty. Flare detection not performed."
		 << Sas::warning("GTI flare rejection"); 
	  return false;
	  
	}


      addFracExp(timebin);

      return true;
    }
  
  ENDPROC
}


void Emproc::addFracExp(double timeBin)
{
  PROC


    /*
     * Retrieve values for the previously calculate rate curve.
     * We will introduce a new column FRACEXP to this eventlist file. This 
     * column gives us the fraction that a bin has been active. 
     */

  DataSet*  tempDS = dataSetServer->open(ePS_->proc.tempFlareBkgTimeseriename(), Dal::Modify); 
  Table* tempTab = tempDS->table("RATE");
    
  double timebin = timeBin;
  //double timeBin;
    double timeStart, timeEnd;
  long nrows;
  
  /*
   * Build full time series with time step of one frame 
   */
  double timeBinSize = 2.6;
  
  
  timeStart = tempTab->attribute("TSTART")->asReal();
  timeEnd = tempTab->attribute("TSTOP")->asReal();
  //timeBin = tempTab->attribute("TIMEDEL")->asReal();
  
  nrows = tempTab->rows();

  /*
   * Call to evselect to calculate the time serie.
   */
  Info::Task* evselect = task("evselect");
  ParameterInfo& evselectpar = *evselect->parameters();
  evselectpar.revertToDefault();
  
  evselectpar["table"]->setValue(ePS_->proc.tempImagingEvlistname());
  evselectpar["maketimecolumn"]->setValue(false);
  evselectpar["makeratecolumn"]->setValue(false);
  evselectpar["timebinsize"]->setValue(timeBinSize);
  evselectpar["timemin"]->setValue(timeStart);
  evselectpar["timemax"]->setValue(timeEnd);
  evselectpar["withrateset"]->setValue(true);
  evselectpar["withtimeranges"]->setValue(true);
  evselectpar["rateset"]->setValue("toto.in.mos");
  
  if(!ePS_->proc.protectedrun(evselect))
    throw(ProcException(""));


  /*
   * Truncate the counts to 1
   */ 
  DataSet* tempRate = dataSetServer->open("toto.in.mos", Dal::Modify);
  
  Table* rateTable = tempRate->table("RATE");

  ColumnData *rateCol = rateTable->column("COUNTS")->data();

  for (unsigned int i = 0; i <  rateTable->column("COUNTS")->rows(); i++)
    rateCol->int32Data()[i] = min(rateCol->int32Data()[i],1);


  /*
   * Rebin into the same bins as the flare timeseries
   */
  int rowCounter = 0;
  double binCounter=0;

  std::vector<double> rebinRate; 

  /*
   * Check that both values are correct
   */

  int newBinSize;

  /*
   * Calculate the new binsize taking into account the number of rows of
   * each time serie: the one we calculated using a time bin of 52 or the number we 
   * obtained from the cheking with the areaCcd and the time serie we have calculated
   * using a time bin of 2.6.
   */
 
  if ( (rateTable->column("COUNTS")->rows() % nrows) == 0 )
      newBinSize = (int)(rateTable->column("COUNTS")->rows() / nrows);
  else
      newBinSize = (int)floor( (double)rateTable->column("COUNTS")->rows()/
			       (double)(nrows)) + 1;
  
  /*
   * Check if the newBinSize is greater than timebin/2.6. This is a way to
   * check that the new rebinning is going to be OK.
   */

  errstr << "Rebining from " <<   rateTable->column("COUNTS")->rows() 
	 << " number of rows to " << nrows << " . New bin size " 
	 << newBinSize << Sas::message(Msg::Meta, Msg::Silent);
  
  if ( (timebin / timeBinSize) > newBinSize)
    errstr << "The new bin size value "  << newBinSize 
	   << " is greater than " << timebin << "/" <<timeBinSize 
	   << " GATTI flaring rejection will be wrong" 
 	   << Sas::warning("addFracExp");
  
  /*
   * Sum up all the count until we reach the new bin size or we reach 
   * the end of the table. Then divide by the bin size to renormalize the 
   * the bin count.
   * Remember that both time series (the one obtained witha a time bin >= 52 and 
   * the time serie obtained with a time bin of 2.6 are correlacted because of this
   * fact.
   */

   
  for (unsigned int i = 0 ; i < rateTable->column("COUNTS")->rows() ; ++i) 
    {
      binCounter +=rateCol->int32Data()[i]; 
      ++rowCounter;
            
      if ( (i != 0) && ( (rowCounter % newBinSize  == 0) ||
			 ( ( rateTable->column("COUNTS")->rows() - 1)  == i) ) )
	{
	  binCounter = binCounter / newBinSize;
	  rebinRate.push_back(binCounter);	  
	  binCounter = 0;
	}
    }
  

  /*
   * Add the FRACEXP to event file
   */
  Column *colFracExp = tempTab->addColumn("FRACEXP",Column::Real32);
  ColumnData *colFracExpData = colFracExp->data();
  real32 *r32Data = colFracExpData->real32Data();
  for (unsigned int i = 0 ; i < rebinRate.size() ; ++i)
   r32Data[i] =  rebinRate[i];
  
  delete rateCol;
  delete colFracExpData;


  /*
   * Convert the rate and error to cts/ks/arcmin2
   */
  
  ColumnData *convertedRateCol = tempTab->column("RATE")->data();
  ColumnData *convertedErrorCol = tempTab->column("ERROR")->data();
  
  for (unsigned int i = 0; i < tempTab->column("RATE")->rows(); i++)
    {
      convertedRateCol->real64Data()[i] = 
	convertedRateCol->real64Data()[i]/r32Data[i]*1000.0/ePS_->areaCcd;
      convertedErrorCol->real64Data()[i] = 
	convertedErrorCol->real64Data()[i]/r32Data[i]*1000.0/ePS_->areaCcd;
    }
  
  /*
   * Write the new units
   */  
  tempTab->column("RATE")->units("cts/ks/arcmin2");
  tempTab->column("ERROR")->units("cts/ks/arcmin2");
  
  delete convertedRateCol;
  delete convertedErrorCol;
  
  /*
   * Add keyword areaccd to remember area value
   */
  *tempTab->addAttribute("IN_FOV", "[arcmin2] Area within the FOV") = ePS_->areaCcd;
  
  /*
   * And a comment for the primary header
   */
  *tempDS->addAttribute("CONTENT")="EPIC FLARE BACKGROUND TIMESERIES";
  
  dataSetServer->remove("toto.in.mos");
  //dataSetServer->close(tempRate);
  dataSetServer->close(tempDS);

  ENDPROC

}


void Emproc::tsToGTI()
{
  PROC
    
  /*
   * Call to tabgtigen
   */
    
  Info::Task* tabgtigen = task("tabgtigen");
  ParameterInfo& tabgtigenpar = *tabgtigen->parameters();
  tabgtigenpar.revertToDefault();

  /*
   * AI: 
   * expression and mingtisize values are taken from the emchain.
   * Maybe, we should have to include this to values as parameter...
   * threshold=2 -> Threshold on singles with truncated E1 (per ks/arcmin2)
   * 78.0 = 1.5 * timebin = 1.5 * 52 
   * timebin = 52 -> Time binning to build histogram for flare rejection
   */


  std::string expression = "'RATE<=2'";

  tabgtigenpar["table"]->setValue(ePS_->proc.tempFlareBkgTimeseriename());
  tabgtigenpar["expression"]->setValue(expression);
  tabgtigenpar["mingtisize"]->setValue(78.0);
  tabgtigenpar["gtiset"]->setValue(ePS_->proc.tempFlareBkgGTIname());
  
  if(!ePS_->proc.protectedrun(tabgtigen))
    throw(ProcException(""));


  ENDPROC
}


/*
 * Check whether flare GTI should be applied or not
 */
void Emproc::testFlareGTI()
{
  PROC
  
  DataSet*  tempDS = dataSetServer->open(ePS_->proc.tempFlareBkgGTIname(), Dal::Read); 
  Table* tempTab = tempDS->table("STDGTI");    
    /*
     * Check if the Flare GTI file has a number of rows greater than 0 
     * If there are GTI...
     */
  
  
  if (tempTab->rows() == 0 )
    {
      errstr << "The GTI file "<< ePS_->proc.badpixfindeventsname()
	     << " is empty " << Sas::warning("GATTI flare screening");
    }
      
    
    /*
     * Sum up the GTI
     */

  ColumnData *startCol = tempTab->column("START")->data();    
  real64 *startColData = startCol->real64Data();

  ColumnData *stopCol = tempTab->column("STOP")->data();    
  real64 *stopColData = stopCol->real64Data();

  double ongti = 0.0;

  for (unsigned int i = 0 ; i <tempTab->rows()  ; ++i)
    ongti = ongti + (stopColData[i] - startColData[i]);
  

  /*
   * Read the ONTIME or ONTIME2 or TELAPSE from the evenlist
   */ 

  DataSet*  eventDS = dataSetServer->open(ePS_->proc.tempImagingEvlistname(), Dal::Read); 
  Table* eventTab = eventDS->table("EVENTS");    

  double ontime = 0.0;

  if (eventTab->hasAttribute("ONTIME") )
    ontime = eventTab->attribute("ONTIME")->asReal();
  else if (eventTab->hasAttribute("ONTIME02") )
    ontime = eventTab->attribute("ONTIME02")->asReal();
  else if (eventTab->hasAttribute("TELAPSE") )
    ontime = eventTab->attribute("TELAPSE")->asReal();
  else 
    {
      errstr << "No ONTIME/ONTIME02/TELASPE found in "
	     << ePS_->proc.tempImagingEvlistname()
	     << " file " << Sas::warning("GATTI flare screening");
    }
  
  delete startCol;
  delete stopCol;

  dataSetServer->close(tempDS);	    
  dataSetServer->close(eventDS);

  /*
   * Check if GTI covers less than 10% of interval.. 
   */
  
  double timebin;
  timebin = ePS_->mypars["timebin"]->realValue();

  if (ongti < 0.1*ontime)
    {
      errstr << "Less than 10% GTI left after standard flare screening: "
	     << 100.0*ongti/ontime 
	     << ". Event list is left unscreened." 
	     << Sas::message(Msg::App, Msg::Verbose);

      ePS_->hasValidGTI = false;
    }
  else if ( ongti > ontime - 0.5*timebin )
    {
      errstr << "No flare was detected. "
	     << Sas::message(Msg::App, Msg::Verbose);

      ePS_->hasValidGTI = false;
    }
  else
    ePS_->hasValidGTI = true;
  

  ENDPROC  
}

void Emproc::combineMergeEvents(ParameterInfo& evlistcombpars)
{
  PROC
    // Change the columns that are propagates by evlistcomb
    // 
      vector<string> emosimgcolnames, emosimgcoltypes;
      emosimgcolnames.push_back("TIME");    emosimgcoltypes.push_back("double");
      emosimgcolnames.push_back("RAWX");    emosimgcoltypes.push_back("int16"); 
      emosimgcolnames.push_back("RAWY");    emosimgcoltypes.push_back("int16"); 
      emosimgcolnames.push_back("DETX");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("DETY");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("X");    emosimgcoltypes.push_back("int32");
      emosimgcolnames.push_back("Y");    emosimgcoltypes.push_back("int32");
      emosimgcolnames.push_back("PHA");     emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("PI");      emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("FLAG");    emosimgcoltypes.push_back("int32");
      emosimgcolnames.push_back("PATTERN"); emosimgcoltypes.push_back("int8");
      emosimgcolnames.push_back("FRAME");    emosimgcoltypes.push_back("int32");
      emosimgcolnames.push_back("ENERGYE1");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("ENERGYE2");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("ENERGYE3");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("ENERGYE4");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("PERIPIX");    emosimgcoltypes.push_back("int8");
      emosimgcolnames.push_back("OFFSETX");    emosimgcoltypes.push_back("int16");
      emosimgcolnames.push_back("OFFSETY");  emosimgcoltypes.push_back("int16"); 

      evlistcombpars["emosimgcolnames"]->setValue(emosimgcolnames);
      evlistcombpars["emosimgcoltypes"]->setValue(emosimgcoltypes);


      vector<string> emostimcolnames, emostimcoltypes;
      emostimcolnames.push_back("TIME");    emostimcoltypes.push_back("double");
      emostimcolnames.push_back("RAWX");    emostimcoltypes.push_back("int16"); 
      emostimcolnames.push_back("PHA");     emostimcoltypes.push_back("int16");
      emostimcolnames.push_back("PI");      emostimcoltypes.push_back("int16");
      emostimcolnames.push_back("FRAME");    emostimcoltypes.push_back("int32");
      emostimcolnames.push_back("RAWY");    emostimcoltypes.push_back("int16"); 
      emostimcolnames.push_back("PERIPIX");    emostimcoltypes.push_back("int8");
      emostimcolnames.push_back("DETX");    emostimcoltypes.push_back("int16");
      emostimcolnames.push_back("DETY");    emostimcoltypes.push_back("int16");
      emostimcolnames.push_back("X");    emostimcoltypes.push_back("int32");
      emostimcolnames.push_back("Y");    emostimcoltypes.push_back("int32");
      emostimcolnames.push_back("OFFSETX");    emostimcoltypes.push_back("int16");

      evlistcombpars["emostimcolnames"]->setValue(emostimcolnames);
      evlistcombpars["emostimcoltypes"]->setValue(emostimcoltypes);

      
  ENDPROC

}

void Emproc::preProcessExposureHook(Proc::ScienceExposure&) {}
void Emproc::finalizeHook(const OdfSet *) {}
void Emproc::screenExposure(const Proc::ScienceExposure& ){}
void Emproc::combineHook(ParameterInfo&) {}
